# AIDEV-NOTE: Core text generation module with deterministic fallback mechanism
# Implements both model-based generation and hash-based deterministic fallback
# AIDEV-NOTE: Fixed fallback behavior - generator now calls
# _deterministic_fallback_generate() when model is None
# AIDEV-NOTE: Added stop sequences integration - DEFAULT_STOP_SEQUENCES
# are now passed to model calls
# AIDEV-NOTE: Fixed determinism issue - now always uses DEFAULT_SEED when
# no explicit seed is provided to ensure consistent outputs
# AIDEV-NOTE: Added generate_iter() method for streaming token generation
# with graceful fallback to word-by-word yielding when model unavailable
# AIDEV-NOTE: Added dynamic model switching (v1.0.0) - generate() and generate_iter()
# now accept model parameters to switch between different models at runtime while
# maintaining deterministic outputs for each model

import hashlib
import os as _os
from typing import Any, Dict, List, Optional, Union, Tuple, Iterator

from ..disk_backed_frecency_cache import DiskBackedFrecencyCache
from ..models.loader import get_generator_model_instance
from ..utils import set_deterministic_environment  # Assuming this is in utils.py
from ..utils import (
    DEFAULT_SEED,
    DEFAULT_STOP_SEQUENCES,
    GENERATION_MAX_NEW_TOKENS,
    LLAMA_CPP_GENERATION_SAMPLING_PARAMS_DETERMINISTIC,
    logger,
    resolve_model_params,
)

# Ensure environment is set for determinism when this module is loaded
set_deterministic_environment(DEFAULT_SEED)


# AIDEV-NOTE: Disk-backed frecency cache for generated text results
# AIDEV-NOTE: Persistent cache for successful generation outputs with configurable limits
# AIDEV-QUESTION: Should fallback results be cached, or only model-generated ones?
# AIDEV-NOTE: Cache capacity and size can be configured via environment variables:
# - STEADYTEXT_GENERATION_CACHE_CAPACITY (default: 256)
# - STEADYTEXT_GENERATION_CACHE_MAX_SIZE_MB (default: 50.0)
_generation_cache = DiskBackedFrecencyCache(
    capacity=int(_os.environ.get("STEADYTEXT_GENERATION_CACHE_CAPACITY", "256")),
    cache_name="generation_cache",
    max_size_mb=float(
        _os.environ.get("STEADYTEXT_GENERATION_CACHE_MAX_SIZE_MB", "50.0")
    ),
)


# AIDEV-NOTE: Main generator class with model instance caching and error handling
# AIDEV-NOTE: Extended to support dynamic model switching via generate() parameters
class DeterministicGenerator:
    def __init__(self):
        self.model = None
        self._logits_enabled = False
        self._current_model_key = "default::default"
        # Load model without logits_all initially
        self._load_model(enable_logits=False)

    def _load_model(
        self,
        enable_logits: bool = False,
        repo_id: Optional[str] = None,
        filename: Optional[str] = None,
        force_reload: bool = False,
    ):
        """Load or reload the model with specific logits configuration.
        
        AIDEV-NOTE: Now supports loading custom models via repo_id and filename.
        """
        self.model = get_generator_model_instance(
            force_reload=force_reload,
            enable_logits=enable_logits,
            repo_id=repo_id,
            filename=filename,
        )
        self._logits_enabled = enable_logits
        self._current_model_key = f"{repo_id or 'default'}::{filename or 'default'}"
        if self.model is None:
            logger.error(
                f"DeterministicGenerator: Model instance is None after attempting to load {self._current_model_key}."
            )

    def generate(
        self,
        prompt: str,
        return_logprobs: bool = False,
        eos_string: str = "[EOS]",
        model: Optional[str] = None,
        model_repo: Optional[str] = None,
        model_filename: Optional[str] = None,
        size: Optional[str] = None,
    ) -> Union[str, Tuple[str, Optional[Dict[str, Any]]]]:
        """Generate text with optional model switching.
        
        Args:
            prompt: Input text prompt
            return_logprobs: Whether to return token log probabilities
            eos_string: End-of-sequence string ("[EOS]" uses model defaults)
            model: Model name from registry (e.g., "qwen2.5-3b")
            model_repo: Custom Hugging Face repository ID
            model_filename: Custom model filename
            size: Size identifier ("small", "medium", "large")
        
        AIDEV-NOTE: Model switching parameters allow using different models without
        restarting. Precedence: model_repo/model_filename > model > size > env vars > defaults.
        """
        # Resolve model parameters
        if model or model_repo or model_filename or size:
            try:
                repo_id, filename = resolve_model_params(model, model_repo, model_filename, size)
            except ValueError as e:
                logger.error(f"Invalid model specification: {e}")
                fallback = _deterministic_fallback_generate(prompt)
                return (fallback, None) if return_logprobs else fallback
        else:
            repo_id = filename = None
        
        # Handle caching only for non-logprobs requests and default model
        if not return_logprobs and repo_id is None and filename is None:
            prompt_str = prompt if isinstance(prompt, str) else str(prompt)
            # Include eos_string in cache key if it's not the default
            cache_key = (
                prompt_str
                if eos_string == "[EOS]"
                else f"{prompt_str}::EOS::{eos_string}"
            )
            cached = _generation_cache.get(cache_key)
            if cached is not None:
                return cached

        if not isinstance(prompt, str):
            logger.error(
                f"DeterministicGenerator.generate: Invalid prompt type: {type(prompt)}. Expected str. Using fallback."
            )
            # Pass string representation to fallback
            fallback = _deterministic_fallback_generate(str(prompt))
            return (fallback, None) if return_logprobs else fallback

        # Check if we need to load a different model
        model_key = f"{repo_id or 'default'}::{filename or 'default'}"
        needs_different_model = model_key != self._current_model_key
        
        # Load appropriate model if needed
        if needs_different_model or (return_logprobs and not self._logits_enabled):
            logger.info(f"Loading model {model_key} with logits={return_logprobs}")
            self._load_model(
                enable_logits=return_logprobs,
                repo_id=repo_id,
                filename=filename,
                force_reload=False,  # Use cache if available
            )

        # AIDEV-NOTE: This is where the fallback to _deterministic_fallback_generate occurs if the model isn't loaded.
        if self.model is None:
            logger.warning(
                "DeterministicGenerator.generate: Model not loaded. "
                "Using fallback generator."
            )
            fallback = _deterministic_fallback_generate(prompt)
            return (fallback, None) if return_logprobs else fallback

        if not prompt or not prompt.strip():  # Check after ensuring prompt is a string
            logger.warning(
                "DeterministicGenerator.generate: Empty or whitespace-only "
                "prompt received. Using fallback generator."
            )
            # Call fallback for empty/whitespace
            fallback = _deterministic_fallback_generate(prompt)
            return (fallback, None) if return_logprobs else fallback

        try:
            # AIDEV-NOTE: Reset model cache before generation to ensure deterministic
            # behavior across multiple calls with the same seed
            if hasattr(self.model, "reset"):
                self.model.reset()

            sampling_params = {**LLAMA_CPP_GENERATION_SAMPLING_PARAMS_DETERMINISTIC}
            # AIDEV-NOTE: Handle custom eos_string - if it's "[EOS]", use default stop sequences
            # Otherwise, add the custom eos_string to stop sequences
            if eos_string == "[EOS]":
                sampling_params["stop"] = DEFAULT_STOP_SEQUENCES
            else:
                # Combine default stop sequences with custom eos_string
                sampling_params["stop"] = DEFAULT_STOP_SEQUENCES + [eos_string]
            # Always use DEFAULT_SEED for determinism
            sampling_params["seed"] = DEFAULT_SEED

            if return_logprobs:
                # Request logprobs for each generated token
                sampling_params["logprobs"] = GENERATION_MAX_NEW_TOKENS

            # AIDEV-NOTE: Use create_chat_completion for model interaction.
            output: Dict[str, Any] = self.model.create_chat_completion(
                messages=[{"role": "user", "content": prompt}], **sampling_params
            )

            generated_text = ""
            logprobs = None
            if output and "choices" in output and len(output["choices"]) > 0:
                choice = output["choices"][0]
                # AIDEV-NOTE: Model response structure for chat completion may vary.
                # Check for 'text' or 'message.content'.
                if "text" in choice and choice["text"] is not None:  # noqa E501
                    generated_text = choice["text"].strip()  # noqa E501
                elif (
                    "message" in choice
                    and "content" in choice["message"]
                    and choice["message"]["content"] is not None
                ):
                    generated_text = choice["message"]["content"].strip()
                if return_logprobs:
                    logprobs = choice.get("logprobs")

            if not generated_text:
                logger.warning(
                    f"DeterministicGenerator.generate: Model returned empty or "
                    f"whitespace-only text for prompt: '{prompt[:50]}...'"
                )

            # Only cache non-logprobs results for default model
            if not return_logprobs and repo_id is None and filename is None:
                prompt_str = prompt if isinstance(prompt, str) else str(prompt)
                # Include eos_string in cache key if it's not the default
                cache_key = (
                    prompt_str
                    if eos_string == "[EOS]"
                    else f"{prompt_str}::EOS::{eos_string}"
                )
                _generation_cache.set(cache_key, generated_text)

            return (generated_text, logprobs) if return_logprobs else generated_text

        except Exception as e:
            logger.error(
                f"DeterministicGenerator.generate: Error during text generation "
                f"for prompt '{prompt[:50]}...': {e}",
                exc_info=True,
            )
            fallback_output = ""
            return (fallback_output, None) if return_logprobs else fallback_output

    def generate_iter(
        self,
        prompt: str,
        eos_string: str = "[EOS]",
        include_logprobs: bool = False,
        model: Optional[str] = None,
        model_repo: Optional[str] = None,
        model_filename: Optional[str] = None,
        size: Optional[str] = None,
    ) -> Iterator[Union[str, Dict[str, Any]]]:
        """Generate text iteratively, yielding tokens as they are produced.

        AIDEV-NOTE: Streaming generation for real-time output. Falls back to
        yielding words from deterministic fallback when model unavailable.

        Args:
            prompt: The input prompt to generate from
            eos_string: Custom end-of-sequence string. "[EOS]" means use model's default.
            include_logprobs: If True, yield dicts with token and logprob info
            model: Model name from registry (e.g., "qwen2.5-3b")
            model_repo: Custom Hugging Face repository ID
            model_filename: Custom model filename
            size: Size identifier ("small", "medium", "large")

        AIDEV-NOTE: When running in pytest, collecting many tokens (>400) can cause
        hanging due to interaction between streaming API and pytest's output capture.
        Tests should limit token collection to avoid this issue. Works fine outside pytest.
        """
        if not isinstance(prompt, str):
            logger.error(
                f"DeterministicGenerator.generate_iter: Invalid prompt type: {type(prompt)}. Expected str. Using fallback."
            )
            # Yield words from fallback
            fallback_text = _deterministic_fallback_generate(str(prompt))
            for word in fallback_text.split():
                yield word + " "
            return

        # Resolve model parameters
        if model or model_repo or model_filename or size:
            try:
                repo_id, filename = resolve_model_params(model, model_repo, model_filename, size)
            except ValueError as e:
                logger.error(f"Invalid model specification: {e}")
                # Yield words from fallback
                fallback_text = _deterministic_fallback_generate(prompt)
                for word in fallback_text.split():
                    yield word + " "
                return
        else:
            repo_id = filename = None
        
        # Check if we need to load a different model
        model_key = f"{repo_id or 'default'}::{filename or 'default'}"
        needs_different_model = model_key != self._current_model_key
        
        # Load appropriate model if needed
        if needs_different_model or (include_logprobs and not self._logits_enabled):
            logger.info(f"Loading model {model_key} with logits={include_logprobs}")
            self._load_model(
                enable_logits=include_logprobs,
                repo_id=repo_id,
                filename=filename,
                force_reload=False,
            )

        # AIDEV-NOTE: Check if model is loaded, fallback to word-by-word generation if not
        if self.model is None:
            logger.warning(
                "DeterministicGenerator.generate_iter: Model not loaded. "
                "Using fallback generator."
            )
            fallback_text = _deterministic_fallback_generate(prompt)
            for word in fallback_text.split():
                yield word + " "
            return

        if not prompt or not prompt.strip():
            logger.warning(
                "DeterministicGenerator.generate_iter: Empty or whitespace-only "
                "prompt received. Using fallback generator."
            )
            fallback_text = _deterministic_fallback_generate(prompt)
            for word in fallback_text.split():
                yield word + " "
            return

        try:
            # AIDEV-NOTE: Reset model cache before generation
            if hasattr(self.model, "reset"):
                self.model.reset()

            sampling_params = {**LLAMA_CPP_GENERATION_SAMPLING_PARAMS_DETERMINISTIC}
            # AIDEV-NOTE: Handle custom eos_string for streaming generation
            if eos_string == "[EOS]":
                sampling_params["stop"] = DEFAULT_STOP_SEQUENCES
            else:
                sampling_params["stop"] = DEFAULT_STOP_SEQUENCES + [eos_string]
            sampling_params["seed"] = DEFAULT_SEED
            sampling_params["stream"] = True  # Enable streaming

            if include_logprobs:
                # Request logprobs for streaming
                sampling_params["logprobs"] = GENERATION_MAX_NEW_TOKENS

            # AIDEV-NOTE: Streaming API returns an iterator of partial outputs
            stream = self.model.create_chat_completion(
                messages=[{"role": "user", "content": prompt}], **sampling_params
            )

            for chunk in stream:
                if chunk and "choices" in chunk and len(chunk["choices"]) > 0:
                    choice = chunk["choices"][0]
                    delta = choice.get("delta", {})

                    # AIDEV-NOTE: Streaming responses use 'delta' for incremental content
                    if include_logprobs:
                        # Yield dict with token and logprob info when requested
                        token_info = {}
                        if "content" in delta and delta["content"]:
                            token_info["token"] = delta["content"]
                        elif "text" in choice and choice["text"]:
                            token_info["token"] = choice["text"]

                        if "logprobs" in choice:
                            token_info["logprobs"] = choice["logprobs"]

                        if "token" in token_info:
                            yield token_info
                    else:
                        # Normal string yielding
                        if "content" in delta and delta["content"]:
                            yield delta["content"]
                        elif "text" in choice and choice["text"]:
                            yield choice["text"]

        except Exception as e:
            logger.error(
                f"DeterministicGenerator.generate_iter: Error during streaming generation "
                f"for prompt '{prompt[:50]}...': {e}",
                exc_info=True,
            )
            # On error, don't yield anything further


# AIDEV-NOTE: Complex hash-based fallback generation algorithm for
# deterministic output when model is unavailable - uses multiple hash seeds
# for word selection
# AIDEV-NOTE: This is the hash-based fallback mechanism.
def _deterministic_fallback_generate(prompt: str) -> str:
    # Ensure prompt_for_hash is always a string, even if original prompt was not.
    if not isinstance(prompt, str) or not prompt.strip():
        prompt_for_hash = f"invalid_prompt_type_or_empty:{type(prompt).__name__}"
        logger.warning(
            f"Fallback generator: Invalid or empty prompt type received "
            f"({type(prompt).__name__}). Using placeholder for hash: "
            f"'{prompt_for_hash}'"
        )
    else:
        prompt_for_hash = prompt

    words = [
        "the",
        "quick",
        "brown",
        "fox",
        "jumps",
        "over",
        "lazy",
        "dog",
        "and",
        "a",
        "in",
        "it",
        "is",
        "to",
        "that",
        "this",
        "was",
        "for",
        "on",
        "at",
        "as",
        "by",
        "an",
        "be",
        "with",
        "if",
        "then",
        "else",
        "alpha",
        "bravo",
        "charlie",
        "delta",
        "echo",
        "foxtrot",
        "golf",
        "hotel",
        "india",
        "juliett",
        "kilo",
        "lima",
        "mike",
        "november",
        "oscar",
        "papa",
        "quebec",
        "romeo",
        "sierra",
        "tango",
        "uniform",
        "victor",
        "whiskey",
        "x-ray",
        "yankee",
        "zulu",
        "error",
        "fallback",
        "deterministic",
        "output",
        "generated",
        "text",
        "response",
        "steady",
        "system",
        "mode",
        "token",
        "sequence",
        "placeholder",
        "content",
        "reliable",
        "consistent",
        "predictable",
        "algorithmic",
        "data",
        "model",
        "layer",
    ]

    hasher = hashlib.sha256(prompt_for_hash.encode("utf-8"))
    hex_digest = hasher.hexdigest()  # Example: '50d858e0985ecc7f60418aaf0cc5ab58...'

    seed1 = int(hex_digest[:8], 16)
    seed2 = int(hex_digest[8:16], 16)
    seed3 = int(hex_digest[16:24], 16)

    try:
        max_tokens_target = GENERATION_MAX_NEW_TOKENS
    except NameError:
        max_tokens_target = 100
        logger.warning("GENERATION_MAX_NEW_TOKENS not found, fallback using 100.")

    num_words_to_generate = (seed3 % 21) + (max_tokens_target - 10)
    num_words_to_generate = max(1, num_words_to_generate)

    fallback_text_parts: List[str] = []

    current_seed = seed1
    for i in range(num_words_to_generate):
        index_val = (current_seed >> (i % 16)) ^ (seed2 + i)
        index = index_val % len(words)
        fallback_text_parts.append(words[index])

        current_seed = (current_seed * 1664525 + seed2 + 1013904223 + i) & 0xFFFFFFFF
        seed2 = (seed2 * 22695477 + current_seed + 1 + i) & 0xFFFFFFFF

    return " ".join(fallback_text_parts)


def _deterministic_fallback_generate_iter(prompt: str) -> Iterator[str]:
    """Iterative version of deterministic fallback that yields words one by one.

    AIDEV-NOTE: Used by generate_iter when model is unavailable. Yields the same
    deterministic output as _deterministic_fallback_generate but word by word.
    """
    fallback_text = _deterministic_fallback_generate(prompt)
    for word in fallback_text.split():
        yield word + " "


# AIDEV-NOTE: Module-level singleton generator instance for backward compatibility
_generator_instance = DeterministicGenerator()


def generate(
    prompt: str,
    return_logprobs: bool = False,
    eos_string: str = "[EOS]",
    model: Optional[str] = None,
    model_repo: Optional[str] = None,
    model_filename: Optional[str] = None,
    size: Optional[str] = None,
) -> Union[str, Tuple[str, Optional[Dict[str, Any]]]]:
    """Generate text deterministically with optional model switching.
    
    This is the main public API for text generation. It maintains backward
    compatibility while adding support for dynamic model switching.
    
    Args:
        prompt: Input text prompt
        return_logprobs: Whether to return token log probabilities
        eos_string: End-of-sequence string ("[EOS]" uses model defaults)
        model: Model name from registry (e.g., "qwen2.5-3b", "qwen3-8b")
        model_repo: Custom Hugging Face repository ID (e.g., "Qwen/Qwen2.5-3B-Instruct-GGUF")
        model_filename: Custom model filename (e.g., "qwen2.5-3b-instruct-q8_0.gguf")
        size: Size identifier ("small", "medium", "large") - maps to Qwen3 0.6B/1.7B/4B models
    
    Returns:
        Generated text string, or tuple of (text, logprobs) if return_logprobs=True
    
    Examples:
        # Use default model (medium/1.7B)
        text = generate("Hello, world!")
        
        # Use size parameter
        text = generate("Quick response", size="small")  # Uses Qwen3-0.6B
        text = generate("Complex task", size="large")    # Uses Qwen3-4B
        
        # Use a model from the registry
        text = generate("Explain quantum computing", model="qwen2.5-3b")
        
        # Use a custom model
        text = generate(
            "Write a poem",
            model_repo="Qwen/Qwen2.5-7B-Instruct-GGUF",
            model_filename="qwen2.5-7b-instruct-q8_0.gguf"
        )
    
    AIDEV-NOTE: Model switching allows using different models without changing
    environment variables. Models are cached after first load for efficiency.
    The size parameter provides convenient access to Qwen3 models of different sizes.
    """
    return _generator_instance.generate(
        prompt=prompt,
        return_logprobs=return_logprobs,
        eos_string=eos_string,
        model=model,
        model_repo=model_repo,
        model_filename=model_filename,
        size=size,
    )


def generate_iter(
    prompt: str,
    eos_string: str = "[EOS]",
    include_logprobs: bool = False,
    model: Optional[str] = None,
    model_repo: Optional[str] = None,
    model_filename: Optional[str] = None,
    size: Optional[str] = None,
) -> Iterator[Union[str, Dict[str, Any]]]:
    """Generate text iteratively with optional model switching.
    
    Yields tokens as they are generated, enabling real-time streaming output.
    
    Args:
        prompt: Input text prompt
        eos_string: End-of-sequence string ("[EOS]" uses model defaults)
        include_logprobs: Whether to include log probabilities in output
        model: Model name from registry (e.g., "qwen2.5-3b")
        model_repo: Custom Hugging Face repository ID
        model_filename: Custom model filename
        size: Size identifier ("small", "medium", "large") - maps to Qwen3 0.6B/1.7B/4B models
    
    Yields:
        String tokens, or dicts with 'token' and 'logprobs' if include_logprobs=True
    
    AIDEV-NOTE: Streaming generation with model switching support. Falls back
    to word-by-word yielding from deterministic fallback when model unavailable.
    """
    return _generator_instance.generate_iter(
        prompt=prompt,
        eos_string=eos_string,
        include_logprobs=include_logprobs,
        model=model,
        model_repo=model_repo,
        model_filename=model_filename,
        size=size,
    )
